from datetime import datetime
from sqlalchemy import select
from sqlalchemy.orm import sessionmaker

from bagels.config import CONFIG
from bagels.models.account import Account
from bagels.models.record import Record
from bagels.models.split import Split
from bagels.models.database.app import db_engine

Session = sessionmaker(bind=db_engine)


def get_account_balance(accountId, session=None):
    """Returns the net balance of an account.

    Rules:
    - Consider all record "account" and split "account"
    - Records with isTransfer should consider both "account" and "transferToAccount"
    - Records and splits should be considered separately, unlike net figures which consider records and splits together.

    Args:
        accountId (int): The ID of the account to get the balance
        session (Session, optional): SQLAlchemy session to use. If None, creates a new session.
    """
    if session is None:
        session = Session()
        should_close = True
    else:
        should_close = False

    try:
        # Initialize balance
        balance = (
            session.query(Account)
            .filter(Account.id == accountId)
            .first()
            .beginningBalance
        )

        # Get all records for this account
        records = session.query(Record).filter(Record.accountId == accountId).all()

        # Calculate balance from records
        for record in records:
            if record.isTransfer:
                # For transfers, subtract full amount (transfers out)
                balance -= record.amount
            elif record.isIncome:
                # For income records, add full amount
                balance += record.amount
            else:
                # For expense records, subtract full amount
                balance -= record.amount

        # Get all records where this account is the transfer destination
        transfer_to_records = (
            session.query(Record)
            .filter(Record.transferToAccountId == accountId, Record.isTransfer == True)
            .all()
        )

        # Add transfers into this account
        for record in transfer_to_records:
            balance += record.amount

        # Get all splits where this account is specified
        splits = session.query(Split).filter(Split.accountId == accountId).all()

        # Add paid splits (they represent money coming into this account)
        for split in splits:
            if split.isPaid:
                if split.record.isIncome:
                    balance -= split.amount
                else:
                    balance += split.amount

        return round(balance, CONFIG.defaults.round_decimals)
    finally:
        if should_close:
            session.close()


def create_account(data):
    session = Session()
    try:
        new_account = Account(**data)
        session.add(new_account)
        session.commit()
        session.refresh(new_account)
        session.expunge(new_account)
        return new_account
    finally:
        session.close()


def _get_base_accounts_query(get_hidden=False):
    stmt = select(Account).filter(Account.deletedAt.is_(None))
    if not get_hidden:
        stmt = stmt.filter(Account.hidden.is_(False))
    else:
        stmt = stmt.order_by(Account.hidden)
    return stmt


def get_all_accounts(get_hidden=False):
    session = Session()
    try:
        stmt = _get_base_accounts_query(get_hidden)
        return session.scalars(stmt).all()
    finally:
        session.close()


def get_accounts_count(get_hidden=False):
    session = Session()
    try:
        stmt = _get_base_accounts_query(get_hidden)
        return len(session.scalars(stmt).all())
    finally:
        session.close()


def get_all_accounts_with_balance(get_hidden=False):
    session = Session()
    try:
        stmt = _get_base_accounts_query(get_hidden)
        accounts = session.scalars(stmt).all()
        for account in accounts:
            account.balance = get_account_balance(account.id, session)
        return accounts
    finally:
        session.close()


def get_account_balance_by_id(account_id):
    session = Session()
    try:
        return get_account_balance(account_id, session)
    finally:
        session.close()


def get_account_by_id(account_id):
    session = Session()
    try:
        return session.get(Account, account_id)
    finally:
        session.close()


def update_account(account_id, data):
    session = Session()
    try:
        account = session.get(Account, account_id)
        if account:
            for key, value in data.items():
                setattr(account, key, value)
            session.commit()
            session.refresh(account)
            session.expunge(account)
        return account
    finally:
        session.close()


def delete_account(account_id):
    session = Session()
    try:
        account = session.get(Account, account_id)
        if account:
            account.deletedAt = datetime.now()
            session.commit()
            return True
        return False
    finally:
        session.close()



import time
import pytest

import os
import time_machine
import datetime as dt
from zoneinfo import ZoneInfo

from textual.pilot import Pilot

hill_valley_tz = ZoneInfo("America/Los_Angeles")

TEMP_INSTANCE_PATH = os.path.join(os.path.dirname(__file__), "../instance/")
SIZE_BASIC = (140, 30)
SIZE_VERTICAL = (70, 50)
SIZE_TEST = (140, 30)
APP_PARAMS = {"is_testing": True}


# ---------- Clear last test --------- #

if os.path.exists(TEMP_INSTANCE_PATH):
    for file in os.listdir(TEMP_INSTANCE_PATH):
        os.remove(os.path.join(TEMP_INSTANCE_PATH, file))

# --------- Init app sequence -------- #
from bagels.locations import set_custom_root

set_custom_root(TEMP_INSTANCE_PATH)

from bagels.config import load_config

load_config()

from bagels.models.database.app import init_db

init_db()

from bagels.app import App

# -------------- Freeze -------------- #


@pytest.fixture(autouse=True)
def travel_to_hill_valley_time():
    with time_machine.travel(dt.datetime(1985, 10, 26, 1, 24, tzinfo=hill_valley_tz)):
        yield  # This allows the tests to run within the context of the travel


# --------------- Basic -------------- #


class TestBasic:
    def test_1_welcome(self, snap_compare):
        assert snap_compare(App(**APP_PARAMS), terminal_size=SIZE_BASIC)

    def test_2_new_acc_welcome(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("a")  # add
            await pilot.press("t", "tab")  # name
            await pilot.press(*"123.45", "enter")  # value

        assert snap_compare(App(**APP_PARAMS), terminal_size=SIZE_BASIC, run_before=r)

    def test_3_vertical_layout(self, snap_compare):
        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_VERTICAL,
        )

    def test_4_jump_screen(self, snap_compare):
        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_BASIC,
            press=["v"],
        )


# ------------- Accounts ------------- #


class TestAccounts:
    def test_5_acc_screen(self, snap_compare):
        assert snap_compare(App(**APP_PARAMS), terminal_size=SIZE_TEST, press=["a"])

    def test_6_new_acc_home(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("v", "a")  # jump to accounts
            await pilot.wait_for_animation()
            await pilot.press("a")  # add
            await pilot.press(*"t2", "enter")  # name

        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )

    def test_7_delete_acc(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("v", "a")  # jump to accounts
            await pilot.press("d", "enter")  # delete

        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )

    def test_8_edit_acc(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("v", "a")  # jump to accounts
            await pilot.press("e")  # edit
            await pilot.press("tab", *"123.234", "enter")  # edit beg. balance

        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )


# ------------ Categories ------------ #


class TestCategories:
    def test_9_categories(self, snap_compare):
        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            press=["c"],
        )

    def test_10_new_category_screen(self, snap_compare):
        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            press=["c", "a"],
        )

    def test_11_new_category(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("c")  # show categories modal
            await pilot.press("a")  # add
            await pilot.press("t", "tab")  # name
            await pilot.press("tab", "tab", "enter")  # nature, color, save

        # todo: scroll to bottom
        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )

    def test_12_new_category_subcategory(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("c")  # show categories modal
            await pilot.press("up")  # select last created
            await pilot.press("s")  # add subcategory
            await pilot.press("t", "tab")  # name
            await pilot.press("tab", "tab", "enter")  # nature, color, save

        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )

    def test_13_delete_category(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("c")  # show categories modal
            await pilot.press("d", "enter")  # delete

        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )


# ------------- Templates ------------ #


class TestTemplates:
    def test_14_templates(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("v", "t")  # jump to templates
            await pilot.press("a")  # add
            await pilot.press("t", "tab")  # name
            await pilot.press("tab")  # category
            await pilot.press(*"123")  # enter value
            await pilot.press("tab")  # account
            await pilot.press("tab", "space", "enter")  # switch to income, save

        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )

    def test_15_edit_template(self, snap_compare):
        async def r(pilot: Pilot):
            await pilot.press("v", "t")  # jump to templates
            await pilot.press("tab")  # select first template
            await pilot.press("e")  # edit
            await pilot.press(*"est")  # edit name
            await pilot.press("enter")  # save

        assert snap_compare(
            App(**APP_PARAMS),
            terminal_size=SIZE_TEST,
            run_before=r,
        )
